#include	"lcd.h"
#include	"p18f4550.h"
#include	"delay.h"

/** 
 * Strobes the Enable control line to trigger the lcd to process the
 * transmitted instruction.
 */
void lcd_strobe_enable(void) {
	LCD_E=1;
	Delay_us(4);
	LCD_E=0;
	Delay_us(4);
}

/**
 * Waits for the busy flag to clear, which should take
 * around the maximum time for an instruction to complete.
 * Note, LCD operation is kind of sensitive to this configuration. If the delay
 * is too fast, the LCD will miss some pixels when it is really put through
 * a stress test. This dela time seems to work great.
 */
void lcd_wait_busy(void) {
	Delay_us(5);
}

/**
 * Older implementation of lcd_wait_busy() that checked the busy flag in 
 * hardware. I found that it always hanged after plotting a byte to the screen,
 * so I took up the delay version above.
 *
void lcd_wait_busy(void) {
	unsigned char counter = 1;
	unsigned char data;
	* Set RW and RS high *
	LCD_RW=1;
	LCD_RS=1;
	Delay_us(4);
	* Wait until busy flag on last bit of the data port clears. *
	do {
		LCD_E=1;
		Delay_us(4);
		data = LCD_DATA_OUT;
		Delay_us(4);
		LCD_E=0;
		Delay_us(4);
	} while (data & 0x80);
}*/

/**
 * Writes a raw instruction to the LCD. 
 * @param command The 4-bit instruction code.
 * @param data The 8-bit paramater/data to the specified instruction.
 */
void lcd_write_command(unsigned char command, unsigned char data) {
	/* Wait for the busy flag to clear */
	lcd_wait_busy();
	
	/* Set RW low, RS high to write the instruction command */
	LCD_RW=0;
	LCD_RS=1;
	/* Instruction commands are a maximum of 4 bits long, so 
	 * just mask off the rest. */
	LCD_DATA_OUT = (command&0x0F);
	Delay_us(5);
	Delay_us(5);
	lcd_strobe_enable();
	Delay_us(5);
	Delay_us(5);

	/* Set RW low, RW low to write the instruction data */
	LCD_RW=0;
	LCD_RS=0;
	LCD_DATA_OUT = data;
	Delay_us(5);
	Delay_us(5);
	lcd_strobe_enable();
	Delay_us(5);
	Delay_us(5);
}

/**
 * Initializes the LCD in graphics mode.
 * Uses a character pitch of 8 (8 bits are plotted whenever a byte is drawn)
 */ 
void lcd_graphics_init(void) {
	unsigned char commandData;
	LCD_RS_TRIS=0;
	LCD_RW_TRIS=0;
	LCD_E_TRIS=0;
	LCD_CS_TRIS=0;
	LCD_RST_TRIS=0;
	LCD_RST=1;
	LCD_CS=0;

	/* Set the data direction registers apprioriately */
	LCD_DATA_DIR = 0x00;

	/* Assert all control lines to low */
	LCD_RW=0;
	LCD_RS=0;
	LCD_E=0;

	/* Send mode configuration command with
	 * Toggle Display On, Master, Mode Graphics bits set */
	commandData = LCD_MODE_ON_OFF | LCD_MODE_MASTER_SLAVE | LCD_MODE_MODE;
	lcd_write_command(LCD_CMD_MODE, commandData);

	/* Send the set character pitch command with horizontal
 	 * character pitch of 8 (so 8 pixels are painted when we draw) */	
	commandData = LCD_CHAR_PITCH_HP_8;
	lcd_write_command(LCD_CMD_CHAR_PITCH, commandData);
	
	/* Send the number of characters command with the total
	 * number of graphics bytes that can be painted horizontally 
	 * (width/8) */
	commandData = (LCD_WIDTH/8)-1;
	lcd_write_command(LCD_CMD_NUM_CHARS, commandData);

	/* Set the time division */
	commandData = 128-1;
	lcd_write_command(LCD_CMD_TIME_DIVISION, commandData);
	
	/* Set the display low/high start address to 0x00 (left corner) */
	commandData = 0x00;
	lcd_write_command(LCD_CMD_DISPLAY_START_LA, commandData);
	lcd_write_command(LCD_CMD_DISPLAY_START_HA, commandData);

	/* Reset the cursor to home 0x00 (left corner) */
	commandData = 0x00;
	lcd_write_command(LCD_CMD_CURSOR_LA, commandData);
	lcd_write_command(LCD_CMD_CURSOR_HA, commandData); 
}

/**
 * Moves the LCD cursor to the specified coordinates. 
 * @param x The new x coordinante of the cursor.
 * @param y The new y coordinante of the cursor.
 */
void lcd_graphics_move(unsigned int x, unsigned int y) {
	unsigned int pos;

	/* Calculate the raw address in terms of bytes on the screen */
	pos = ((y*LCD_WIDTH)+x)/8;

	/* Move the cursor to the new address */
	lcd_write_command(LCD_CMD_CURSOR_LA, pos&0xFF);
	lcd_write_command(LCD_CMD_CURSOR_HA, pos>>8);
}

/**
 * Draws a byte to the LCD at the current LCD's cursor location.
 * @param data The byte to draw. The pixels are drawn MSB to LSB.
 */
void lcd_graphics_draw_byte(unsigned char data) {
	lcd_write_command(LCD_CMD_WRITE_DATA, data);
}

/**
 * Plots a byte at the specified coordinates. 
 * @param x The x coordinante of the byte to be drawn.
 * @param y The y coordinante of the byte to be drawn.
 * @param data The byte to draw. The pixels are drawn MSB to LSB.
 */
void lcd_graphics_plot_byte(unsigned int x, unsigned int y, unsigned char data) {
	lcd_graphics_move(x, y);
	lcd_graphics_draw_byte(data);
}

/**
 * Plots a pixel at the specified coordinates.
 * @param x The x coordinante of the pixel.
 * @param y The y coordinante of the pixel.
 * @param state PIXEL_ON to set the pixel, otherwise pixel will be cleared.
 */
void lcd_graphics_plot_pixel(unsigned int x, unsigned int y, unsigned char state) {
	unsigned char pos;
	
	lcd_graphics_move(x, y);
	/* Since lcd_graphics_move() moves the cursor to a particular
	 * byte, not bit, we need the relative distance to the specified 
	 * bit we are going to set/clear. */
	pos = x%8;

	if (state == PIXEL_ON) 
		lcd_write_command(LCD_CMD_SET_BIT, pos);
	else 
		lcd_write_command(LCD_CMD_CLEAR_BIT, pos);
}

/**
 * Clears the LCD screen 
 */
void lcd_graphics_clear(void) {
	unsigned int i;
	/* Move cursor to home (top left corner) */
	lcd_graphics_move(0, 0);
	/* Draw empty bytes to ocucpy the entire screen */
	for (i = 0; i < ((LCD_WIDTH*LCD_HEIGHT)/8); i++) 
		lcd_graphics_draw_byte(0x00);
}
void lcd_test_clear(void) {
	unsigned int i;
	/* Move cursor to home (top left corner) */
	lcd_graphics_move(0, 0);
	/* Draw empty bytes to ocucpy the entire screen */
	for (i = 0; i < 320; i++) 
		lcd_graphics_draw_byte(' ');
}
/**
 * Delay in milliseconds.
 * An extension of _delay_ms() so we can delay for longer periods of time.
 * @param ms Milliseconds to delay.
 */
void delay_ms_long(unsigned int ms) {
	for (; ms > 0; ms--)
	Delay_ms(1);
} 